/*
 * $RCSfile: event.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990,1996  Arizona Board of Regents
 *
 * $Log: event.c,v $
 * Revision 1.7  1997/03/10 21:37:34  mjk
 * *** empty log message ***
 *
 * Revision 1.6  1996/11/04 22:36:56  mjk
 * stdio is no buffered
 * thread priorities optimized
 * stack warnings correct for events
 *
 * Revision 1.5  1996/09/27 21:23:28  mjk
 * added config.h (needed for prev change).
 *
 * Revision 1.4  1996/09/27 20:39:28  mjk
 * fixed thread stack size warnings.
 *
 * Revision 1.3  1996/03/15 22:40:07  mjk
 * timespec in now POSIX
 *
 * Revision 1.2  1996/03/11  18:02:23  mjk
 * Udptest passes w/ Linux 1.2.13 and Pthreads.  All signals are caught by
 * specific thread(s).
 *
 * Revision 1.1  1996/02/19  20:21:59  mjk
 * Initial revision
 *
 */


#include <signal.h>
#include <errno.h>
#include <time.h>

#include "thread.h"
#include "xk_debug.h"
#include "upi.h"
#include "platform.h"
#include "x_util.h"
#include "event.h"
#include "event_i.h"
#include "xk_assert.h"
#include "x_libc.h"
#include "config.h"

typedef int (*intFunc)();
     
#define THIS_IS_THE_HEADER ((EvFunc)-42)


/* BIG_N should be a function of the estimated number of scheduled events,
 * such that queues don't get too long, but with as little overhead as
 * possible. 128 is probably good for up to some 500-1000 events. [mats]
 */

#define BIG_N 128 

static struct Event 	evHead[BIG_N];

static Event    evSelf( void );


int     eventStackWarnLimit  = EVENT_STACK_SIZE - 0x2000;
int     eventStackErrorLimit = EVENT_STACK_SIZE - 0x0500;


#define THREAD_T	pthread_t
#define THREAD_SELF()	pthread_self() 
#define	EVENT_LOCK()	pthread_mutex_lock(&event_mutex)
#define EVENT_UNLOCK() 	pthread_mutex_unlock(&event_mutex)

pthread_mutex_t	        event_mutex;
static int 		tickmod;
static int event_interval;

#ifdef XK_THREAD_TRACE

Map		localEventMap;
static Map	threadEventMap;

#endif


void evClock(void *unused);		/* Forward */

/*
 * evInit -- initialize the timer queues and start the clock thread
 *
 */
void
evInit(int interval)
{
  int 			i;
  pthread_t 		child;
  static pthread_attr_t	attrs;
  struct sched_param	param;
  sigset_t		set;
  
  /*
   * this must be called from the main thread, so that
   * all other threads will also block SIGALRM
   */
  sigemptyset(&set);
  sigaddset(&set,SIGALRM);
  pthread_sigmask(SIG_BLOCK, &set, NULL);

  
  event_interval=interval;
  xTrace0(event,TR_GROSS_EVENTS,"evInit enter");

#ifdef XK_THREAD_TRACE
  localEventMap = mapCreate(BIG_N, sizeof(Event));    
  threadEventMap = mapCreate(BIG_N, sizeof(THREAD_T));
#endif
  for(i=0; i<BIG_N; i++)
  {
    evHead[i].next = &evHead[i];
    evHead[i].prev = &evHead[i];
    evHead[i].func = THIS_IS_THE_HEADER;
  }
  tickmod = 0;

  pthread_mutex_init(&event_mutex, NULL);

	  /* Since evClock does not use any xkernel routines, this
	     should be allowed to remain a pthread, in spite of the
	     sledgehammer concurrency control otherwise enforced in
	     the xkernel. */

  xTrace0(event, TR_EVENTS, "evInit starting evClock thread");
  pthread_attr_init(&attrs);
  pthread_attr_setdetachstate(&attrs, 1);
  pthread_attr_getschedparam(&attrs, &param);
				/* This statement is NOT POSIX both
                                   the field name and symbol are
                                   nonportable. */
  param.sched_priority = THREAD_MINPRIO; /* this might be too slow 		*/
  pthread_attr_setschedparam(&attrs, &param);
  pthread_create(&child, &attrs, (void *(*)(void *)) &evClock, 0);
  pthread_attr_destroy(&attrs);
    
  xTrace0(event,TR_EVENTS,"evInit exit");

}

/*
 *  e_insque  -- insert an event into an event bucket, ordered by
 *               increasing timeout value
 */

static void
e_insque(int q, Event e)
{
  Event a;

  xTrace0(event,TR_FULL_TRACE,"e_insque enter");
  EVENT_LOCK();
  for (a = evHead[q].next; a != &evHead[q]; a = a->next) {
    if (a->deltat < e->deltat) {
      /* E goes after a */
      e->deltat -= a->deltat;
      continue;
    } else {
      /* E goes just before a */
      a->deltat -= e->deltat;
      insque((struct qelem *)e, (struct qelem *)a->prev);
      EVENT_UNLOCK();
      return;
    }
  }
	  /* E goes at the end */
  insque((struct qelem *)e, (struct qelem *)evHead[q].prev);
  EVENT_UNLOCK();
}

/*
 * e_remque -- remove an event from an event bucket
 */
static void
e_remque(Event e)
{
  xTrace0(event,TR_FULL_TRACE,"e_remque enter");
  xAssert(e);
  if (e->next->func != THIS_IS_THE_HEADER) {
    e->next->deltat += e->deltat;
  }
  remque((struct qelem *)e);
}

#ifdef XK_THREAD_TRACE

/* 
 * It is safe to call unbindEvent multiple times with the same event.
 * Only the first call will have any effect.
 */
#define unbindEvent( _e )                                       \
do {                                                            \
    xAssert(_e);                                                \
    if ( (_e)->bind ) {                                         \
        mapRemoveBinding(threadEventMap, (_e)->bind);           \
        (_e)->bind = 0;                                         \
    }                                                           \
} while(0)

#define freeEvent( _e )                                 \
do {                                                    \
    xAssert(_e);                                        \
    unbindEvent(_e);                                    \
    mapUnbind(localEventMap, &(_e));                    \
    xFree((char *)(_e));                                \
} while (0)

#else	!  XK_THREAD_TRACE 

#define unbindEvent( _e ) do { } while (0)
#define freeEvent(_e)   xFree((char *)(_e))

#endif XK_THREAD_TRACE



static Event
evSelf(void)
{
#ifdef XK_THREAD_TRACE
  Event       ev;
  THREAD_T    thSelf;

  thSelf = THREAD_SELF();
  if ( mapResolve(threadEventMap, &thSelf, &ev) == XK_SUCCESS ) {
    return ev;
  } else {
    xTrace0(event, TR_EVENTS, "evSelf ... couldn't find event structure");
    return 0;
  }
#else
  return 0;
#endif
}



void
evMarkBlocked(Semaphore *s)
{
#ifdef XK_THREAD_TRACE
  Event       ev;

  ev = evSelf();
  if ( ev ) {
    ev->state = E_BLOCKED;
    xGetTime(&ev->stopTime);
  }
#endif
}



void
evMarkRunning(void)
{
#ifdef XK_THREAD_TRACE
  Event       ev;

  ev = evSelf();
  if ( ev ) {
    ev->state = E_RUNNING;
  }
#endif
}


/*
 *  stub -- all scheduled events start here
 *
 *  CreateProcess will invoke stub, and stub will apply the function
 *  pointer to the event object and the argument
 *
 */

static void
stub(Event e)
{
  xTrace0(event,TR_FULL_TRACE,"event stub enter");
  xAssert( e->state == E_SCHEDULED );
#ifdef XK_THREAD_TRACE
  xAssert( mapResolve(localEventMap, &e, 0) == XK_SUCCESS );
#endif

  if ( e->flags & E_CANCELLED_F ) {
    xTrace1(event, TR_MORE_EVENTS,
	    "event stub not starting cancelled event %x", e);
    freeEvent(e);
    return;
  }
#ifdef XK_THREAD_TRACE
  {
    THREAD_T		self;
	
    self = THREAD_SELF();
    e->bind = mapBind(threadEventMap, &self, e);
    e->earlyStack = &self;
    xTrace1(event, TR_MORE_EVENTS, "initial event stack == %x",
	    (int)e->earlyStack);
    xAssert( e->bind != ERR_BIND );
    xGetTime(&e->startTime);
  }
#endif
  e->state = E_RUNNING;
  xTrace1(event,TR_FULL_TRACE, "starting event at addr %x", e->func);
  e->func(e, e->arg);
  e->state = E_FINISHED;
  EVENT_LOCK();
  if ( e->flags & E_DETACHED_F ) {
    freeEvent(e);
  } else {
    unbindEvent(e);
#ifdef XK_THREAD_TRACE
    xGetTime(&e->stopTime);
#endif
  }
  EVENT_UNLOCK(); 
}
      

/*
 *  evSchedule  -- prepare an event for later scheduling
 *
 *
 */

Event
evSchedule( EvFunc func, VOID *arg, unsigned etime)  /* time in usec */
{
  Event e;
  u_int delt;
  bool ret;

  xTrace0(event,TR_FULL_TRACE,"evSchedule enter");
  xTrace3(event, TR_GROSS_EVENTS, "evSchedule: f = %x, arg = %x, etime = %d",
	  func, arg, etime);
  e = (Event)xMalloc(sizeof(struct Event));
  xAssert(e);
  e->func = func;
  e->arg = arg;

#ifdef XK_THREAD_TRACE
  {
    XTime	offset; 
    Bind	b;

    offset.sec = etime / 1E6;
    offset.usec = etime % (unsigned)1E6;
	
      
    xAssert(localEventMap); 
    b = mapBind(localEventMap, &e, e);
    xAssert( b != ERR_BIND );
    e->bind = 0;   /* filled in in 'stub' once a native thread is assigned */
    xGetTime(&e->startTime);
    xTrace2(event, TR_DETAILED, "event start time: %d %d",
	    e->startTime.sec, e->startTime.usec);
    xAddTime(&e->startTime, e->startTime, offset);
  }
#endif
  
	  /*
	   * event_interval, time between scans of the event list, is also in
	   * usec.
	   */
  delt = (etime + event_interval / 2) / event_interval;
  if (delt == 0 && etime != 0) {
    delt = 1;
  }
  xTrace1(event, TR_EVENTS, "event requires %d clock ticks", delt);
  e->deltat = delt / BIG_N + 1;
  e->flags = 0;

  if (delt == 0) {
    e->state = E_SCHEDULED;

    xTrace0(event,TR_EVENTS,"evSchedule starting event");
    ret = CreateProcess((void *)&stub,THREAD_MAXPRIO,1,e);
  } 
  else {
    xTrace1(event, TR_FUNCTIONAL_TRACE, "Placing event in queue %d", (tickmod+delt)%BIG_N);
	    /* e_insque will take care of locking */
    e->state = E_PENDING;
    e_insque((tickmod + delt) % BIG_N, e);
  }
  xTrace1(event, TR_MAJOR_EVENTS, "evSchedule returns event %x", e);
  return e;
}


/*
 * evDetach -- free the storage used by event, or mark it
 *             as detachable after the associate thread finishes
 *
 */

void
evDetach(Event e)
{
  xTrace1(event, TR_GROSS_EVENTS, "evDetach: event = %x", e);
  EVENT_LOCK();
  if ( e->state == E_FINISHED ) {
    freeEvent(e);
  } else {
    e->flags |= E_DETACHED_F;
  }
  EVENT_UNLOCK();
}


/* 
 * See description in event.h
 */
EvCancelReturn
evCancel(Event e)
{
  int ans = 0;

  xTrace1(event, TR_GROSS_EVENTS, "evCancel: event = %x", e);
  xAssert(e);
  EVENT_LOCK();
#ifdef XK_DEBUG
  {
    xkern_return_t	xkr;

    xkr = mapResolve(localEventMap, &e, 0);
    if( xkr != XK_SUCCESS ) {
      Kabort("Attempt to cancel nonexistent event");
    }
  }
#endif

  e->flags |= ( E_DETACHED_F | E_CANCELLED_F );

  switch ( e->state ) {

    case E_SCHEDULED:
	      /* 
	       * The stub routine will catch this event before it gets a chance to
	       * run
	       */
      xTrace2(event, TR_MORE_EVENTS, "cancelled event %x in SCHEDULED state, flags %x",e, e->flags);
      ans = EVENT_CANCELLED;
      break;

    case E_BLOCKED:
    case E_RUNNING:
      ans = EVENT_RUNNING;
      break;

    case E_FINISHED:
      ans = EVENT_FINISHED;
      freeEvent(e);
      break;

    case E_PENDING:
      e_remque(e);
      ans = EVENT_CANCELLED;
      freeEvent(e);
      break;

    default:
      xError("broken event state in evCancel");
  }
  EVENT_UNLOCK();
  return ans;
}
  

int
evIsCancelled(Event e)
{
  int res;

  xAssert(e);
  EVENT_LOCK();
  res = e->flags & E_CANCELLED_F;
  EVENT_UNLOCK();
  return res;
}

void
evCheckStack(char *name)
{
#ifdef XK_THREAD_TRACE
    Event       ev;
    int         diff;

    ev = evSelf();
    if ( ev ) {
        diff = (char *)&ev - (char *)ev->earlyStack;
        if ( diff < 0 ) diff = -diff;
        if ( diff > eventStackWarnLimit ) {
            if ( traceevent >= TR_SOFT_ERRORS || diff > eventStackErrorLimit ) {
                xError("x-kernel stack warning");
                xTrace2(event, TR_ALWAYS,
                        "x-kernel stack warning [%s], event %x", name, ev);
                xTrace3(event, TR_ALWAYS,
                        "cur = %x, beg = %x, size = %d",
                        (int)&ev, (int)ev->earlyStack, diff);
            }
        }
    }
#endif
}

/*
 * evClock --- update the event queues periodically
 *
 * This routine is not under the master lock.
 * It decrements the tick count for each event bucket; events
 * whose time has expired are scheduled for execution via the
 * master lock.  The events are stored in the bucket in order
 * of increasing time to expiration.  Each event's "deltat" field
 * indicates the time by which it exceeds the previous event's
 * timeout value.
 *
 */

void evClock(void *unused)
{
  Event e;
  bool  ret;

	  /*  wait for a clock tick */
  while (1)
  {
    struct timespec	reqnap;
    struct timespec	remnap;
    
    reqnap.tv_sec  = 0;
    reqnap.tv_nsec = event_interval * 1000;
    while ( xnanosleep(&reqnap, &remnap) )
      reqnap = remnap;

	    /* do some work */
/*     xTrace0(event,TR_EVENTS,"evClock tick"); */

    EVENT_LOCK();
    evHead[tickmod].next->deltat--;
    tickmod = (tickmod+1)%BIG_N; 

/*     evDump(); */
    
    for (e = evHead[tickmod].next;
	 e != &evHead[tickmod] && e->deltat == 0; e = e->next)
    {

      remque((struct qelem *)e);
      e->state = E_SCHEDULED;

#ifdef XK_THREAD_TRACE
      xGetTime(&e->startTime);
#endif

      xTrace0(event,TR_EVENTS,"evClock starting event");
      ret = CreateProcess((void *)&stub,THREAD_STDPRIO, 1, e);
    }
    EVENT_UNLOCK();
  }
}



