/* 
 * $RCSfile: fddi.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: fddi.c,v $
 * Revision 1.5  1994/02/05 00:05:29  menze
 *   [ 1994/01/30          menze ]
 *   Using xk_u_intN types
 *
 * Revision 1.4  1994/01/10  17:50:25  menze
 *   [ 1993/12/23          menze ]
 *   Revised rom parsing code to use library routines
 *
 *   [ 1993/12/23          menze ]
 *   Removed irix-platform-specific coding conventions so this protocol can
 *   be built everywhere.
 *
 * Revision 1.3  1993/12/16  01:27:51  menze
 * includes mac.h
 *
 * Revision 1.2  1993/11/13  00:52:47  menze
 * Original version from UMass
 *
 * Based on xkernel/protocols/eth/eth.c (DJY)
 */

/*
 * The xkernel FDDI driver is structured in two layers.
 *
 * The FDDI protocol layer (this file) is independent of any
 * particular FDDI controller hardware.
 * It comprises the usual xkernel protocol functions,
 * e.g. fddiOpen, fddiPush, etc.).
 * It knows about FDDI 48 bit addresses, IEEE 802.2 LLC, and "types,"
 * but nothing about any particular FDDI controller.
 *
 * The device driver, which exports an xkernel interface, sits below
 * this protocol
 */

#include "xkernel.h"
#include "romopt.h"
#include "fddi.h"
#include "fddi_i.h"
#include "mac.h"

typedef struct {
    FDDIhdr	hdr;
} SState;

typedef struct {
    FDDIhost    host;          /* remote host MAC address */
    ETHtype	type;          /* Ethernet type field in 802.2 header */
} ActiveId;

typedef struct {
    XObj    	prmSessn;
    FDDIhost    myHost;        /* our local MAC address */
    xk_u_int8	dsap;          /* SNAP destination service access point */
    xk_u_int8	ssap;          /* SNAP source service access point */
    xk_u_int8   control;       /* SNAP control field */
    xk_u_int8   OrgCodePart1;  /* SNAP Organization code (3 bytes) ... */
    xk_u_int16  OrgCodePart2;
    Map		activeMap;
    Map 	passiveMap;
    int		mtu;
} PState;

typedef ETHtype  PassiveId;  /* Ethernet type field in 802.2 header */

typedef struct {
    Msg		msg;
    XObj	self;
    XObj	llp;
    FDDIhdr	hdr;
} RetBlock;

/*
 * Some nice prime numbers for hashing.
 */
#define	FDDI_ACTIVE_MAP_SIZE	257
#define	FDDI_PASSIVE_MAP_SIZE	13

int 	tracefddip;

#ifdef XK_DEBUG

static FDDIhost	fddiBroadcastHost = FDDI_BCAST_HOST;

#endif


#ifdef __STDC__

static void		demuxStub( Event, VOID * ); 
static int		dispActiveMap( VOID *, VOID *, VOID * );
static int		dispPassiveMap( VOID *, VOID *, VOID * );
static XObj		fddiCreateSessn( XObj, XObj, XObj, ActiveId * );
static void		fddiSessnInit( XObj );
static int		fddiControlProtl( XObj, int, char *, int );
static xkern_return_t	fddiDemux( XObj, XObj, Msg * );
static XObj		fddiOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	fddiOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t	fddiOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t	fddiOpenDisableAll( XObj, XObj );
static xkern_return_t	fddiClose( XObj );
static xmsg_handle_t	fddiPush( XObj, Msg * );
static xmsg_handle_t	fddiLoopPush( XObj, Msg * );
static xkern_return_t	fddiPop( XObj, XObj, Msg *, VOID * );
static int		fddiControlSessn( XObj, int, char *, int );
static long		getRelProtNum( XObj, XObj, char * );
static xkern_return_t	readMtu( XObj, char **, int, int, VOID * );

#else

static void		demuxStub(); 
static int		dispActiveMap();
static int		dispPassiveMap();
static XObj		fddiCreateSessn();
static void		fddiSessnInit();
static int		fddiControlProtl();
static xkern_return_t	fddiDemux();
static XObj		fddiOpen();
static xkern_return_t	fddiOpenEnable();
static xkern_return_t	fddiOpenDisable();
static xkern_return_t	fddiOpenDisableAll();
static xkern_return_t	fddiClose();
static xmsg_handle_t	fddiPush();
static xmsg_handle_t	fddiLoopPush();
static xkern_return_t	fddiPop();
static int		fddiControlSessn();
static long		getRelProtNum();
static xkern_return_t	readMtu();

#endif /* __STDC__ */


#define FDDI_ADDRESS_ISBROAD(A) (FDDI_ADDRS_EQUAL(A,fddiBroadcastHost))
#define FDDI_ADDRESS_ISMULTI(A) (FDDI_ADDRESS_ISGROUP(A) && \
                                 !FDDI_ADDRESS_ISBROAD(A))


static XObjRomOpt fddiOpt[] = {
    { "mtu", 3, readMtu }
};


static long
getRelProtNum( hlp, llp, s )
    XObj	hlp, llp;
    char	*s;
{
    long	n;

    n = relProtNum(hlp, llp);
    if ( n == -1 ) {
	xTrace3(fddip, TR_ERRORS,
		"fddi %s could not get prot num of %s relative to %s",
		s, hlp->name, llp->name);
    }
    if ( n < 0 || n > 0xffff ) {
	return -1;
    }
    return n;
}


static xkern_return_t
readMtu( self, str, nFields, line, arg )
    XObj	self;
    int		line, nFields;
    char	**str;
    VOID	*arg;
{
    PState	*ps = (PState *)self->state;

    return
#ifdef XKMACHKERNEL
				    sscanf1
#else
				    sscanf
#endif
				      (str[2], "%d", &ps->mtu) < 1  ?
  						XK_FAILURE : XK_SUCCESS;
}


void
fddi_init( self )   /* Changed from fddi_init for first build */
     XObj self;
{
    PState	*ps;
    XObj	llp;

    xTrace0(fddip, TR_EVENTS, "fddi_init");
    if ( !xIsProtocol(llp = xGetDown(self, 0)) ) {
	xError("FDDI can not get driver protocol object");
	return;
    }
    if ( xOpenEnable(self, self, llp, 0) == XK_FAILURE ) {
	xError("FDDI can not openenable driver protocol");
	return;
    }
    ps = X_NEW(PState);
    self->state = (VOID *)ps;
    ps->activeMap = mapCreate(FDDI_ACTIVE_MAP_SIZE, sizeof(ActiveId));
    ps->passiveMap = mapCreate(FDDI_PASSIVE_MAP_SIZE, sizeof(PassiveId));
    ps->prmSessn = 0;
    ps->dsap = FDDI_SNAP_DSAP;
    ps->ssap = FDDI_SNAP_SSAP;
    ps->control = FDDI_SNAP_CONTROL;     
    ps->OrgCodePart1 = FDDI_SNAP_ORGCODEPART1;
    ps->OrgCodePart2 = FDDI_SNAP_ORGCODEPART2;
    if( xControl(llp, GETMAXPACKET, (char *)&ps->mtu, sizeof(ps->mtu))
         != sizeof(ps->mtu) ) {
        xTrace0(fddip, TR_ERRORS, "FDDI can't get MTU from driver protocol");
        Kabort("fddi_init: out of here");
    }
    findXObjRomOpts(self, fddiOpt, sizeof(fddiOpt)/sizeof(XObjRomOpt), 0);
    xTrace1(fddip, TR_MAJOR_EVENTS, "FDDI using mtu %d", ps->mtu);
    /*
     * NO ...
     * if ( xControl(llp,MAC_REGISTER_ULP,(char *)&self,sizeof(XObj)) < 0 ) {
     *   xError("fddi_init: can't get register myself with driver protocol");
     *   return;
     * }
     */
    if ( xControl(llp, GETMYHOST, (char *)&ps->myHost, sizeof(FDDIhost))
			< (int)sizeof(FDDIhost) ) {
	xError("fddi_init: can't get my own host");
	return;
    }
    self->control = fddiControlProtl;
    self->open = fddiOpen;
    self->openenable = fddiOpenEnable;
    self->opendisable = fddiOpenDisable;
    self->demux = fddiDemux;
    self->opendisableall = fddiOpenDisableAll;
}


static XObj
fddiOpen( self, hlpRcv, hlpType, part )
    XObj	self, hlpRcv, hlpType;
    Part 	*part;
{
    PState	*ps = (PState *)self->state;
    ActiveId  	key;
    XObj 	fddiSessn;
    FDDIhost	*remoteHost;
    long	protNum;
    
    if ( part == 0 || partLen(part) < 1 ) {
	xTrace0(fddip, TR_SOFT_ERRORS, "fddiOpen - bad participants");
	return ERR_XOBJ;
    }
    remoteHost = (FDDIhost *)partPop(*part);
    xAssert(remoteHost);
    key.host = *remoteHost;
    if ( (protNum = getRelProtNum(hlpType, self, "open")) == -1 ) {
	return ERR_XOBJ;
    }
    key.type = protNum;
    xTrace2(fddip, TR_MAJOR_EVENTS, "fddi_open: destination address = %s:%4x",
	    fddiHostStr(&key.host), key.type);
    key.type = htons(key.type);
    if ( mapResolve(ps->activeMap, &key, &fddiSessn) == XK_FAILURE ) {
	fddiSessn = fddiCreateSessn(self, hlpRcv, hlpType, &key);
    }
    xTrace1(fddip, TR_MAJOR_EVENTS, "fddi_open: returning %X", fddiSessn);
    return (fddiSessn);

}


static XObj
fddiCreateSessn( self, hlpRcv, hlpType, key )
    XObj	self, hlpRcv, hlpType;
    ActiveId	*key;
{
    XObj	s;
    XObj	llp = xGetDown(self, 0);
    SState	*ss;
    PState	*ps = (PState *)self->state;

    s = xCreateSessn(fddiSessnInit, hlpRcv, hlpType, self, 1, &llp);
    if ( FDDI_ADDRS_EQUAL(key->host, ps->myHost) ) {
	xTrace0(fddip, TR_MAJOR_EVENTS,
		"fddiCreateSessn -- creating loopback session");
	s->push = fddiLoopPush;
    }
    s->binding = (Bind) mapBind(ps->activeMap, key, s);
    if ( s->binding == ERR_BIND ) {
	xTrace0(fddip, TR_ERRORS, "error binding in fddiCreateSessn");
	xDestroy(s);
	return ERR_XOBJ;
    }
    ss = X_NEW(SState);
    ss->hdr.dst = key->host;
    ss->hdr.src = ps->myHost;
    ss->hdr.llc.dsap = ps->dsap;
    ss->hdr.llc.ssap = ps->ssap;
    ss->hdr.llc.control = ps->control;
    ss->hdr.llc.OrgCodePart1 = ps->OrgCodePart1;
    ss->hdr.llc.OrgCodePart2 = ps->OrgCodePart2;
    ss->hdr.llc.EtherType = key->type;
    s->state = (VOID *)ss;
    return s;
}


static xkern_return_t
fddiOpenEnable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    PassiveId	key;
    long	protNum;
    
    if ( (protNum = getRelProtNum(hlpType, self, "openEnable")) == -1 ) {
	return XK_FAILURE;
    }
    xTrace2(fddip, TR_GROSS_EVENTS, "fddiOpenEnable: hlp=%x, protlNum=%x",
	    hlpRcv, protNum);
    key = protNum;
    key = htons(key);
    return defaultOpenEnable(ps->passiveMap, hlpRcv, hlpType, (VOID *)&key);
} 


static xkern_return_t
fddiOpenDisable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    long	protNum;
    PassiveId	key;
    
    if ( (protNum = getRelProtNum(hlpType, self, "opendisable")) == -1 ) {
	return XK_FAILURE;
    }
    xTrace2(fddip, TR_GROSS_EVENTS, "fddiOpenDisable: hlp=%x, protlNum=%x",
	    hlpRcv, protNum);
    key = protNum;
    key = htons(key);
    return defaultOpenDisable(ps->passiveMap, hlpRcv, hlpType, (VOID *)&key);
}


static int
dispActiveMap( key, val, arg )
    VOID *        key;
    VOID *        val;
    VOID *        arg;
{
    XObj	s = (XObj)val;
    xPrintXObj(s);
    return MFE_CONTINUE;
}


static int
dispPassiveMap( key, val, arg )
    VOID *         key;
    VOID *         val;
    VOID *         arg;
{
#ifdef XK_DEBUG
    Enable	*e = (Enable *)val;
#endif
    xTrace2(fddip, TR_ALWAYS, "Enable object, hlpRcv == %s, hlpType = %s",
	    e->hlpRcv->fullName, e->hlpType->fullName);
    return MFE_CONTINUE;
}


static xkern_return_t
fddiOpenDisableAll( self, hlpRcv )
    XObj	self, hlpRcv;
{
    xkern_return_t	xkr;
    PState		*ps = (PState *)self->state;

    xTrace0(fddip, TR_MAJOR_EVENTS, "fddiOpenDisableAll called");

    xTrace0(fddip, TR_ALWAYS, "before passive map contents:");
    mapForEach(ps->passiveMap, dispPassiveMap, 0);
    xkr = defaultOpenDisableAll(((PState *)self->state)->passiveMap,hlpRcv,0);
    xTrace0(fddip, TR_ALWAYS, "after passive map contents:");
    mapForEach(ps->passiveMap, dispPassiveMap, 0);
    xTrace0(fddip, TR_ALWAYS, "active map contents:");
    mapForEach(ps->activeMap, dispActiveMap, 0);
    return XK_SUCCESS;
}


static xkern_return_t
fddiDemux( self, llp, msg )
    XObj	self, llp;
    Msg		*msg;
{
    PState	*ps = (PState *)self->state;
    ActiveId 	actKey;
    PassiveId 	pasKey;
    XObj 	s = 0;
    Enable 	*e;
    FDDIhdr	*hdr = msgGetAttr(msg, 0);
    
    xTrace0(fddip, TR_EVENTS, "fddiDemux");
    xTrace1(fddip, TR_FUNCTIONAL_TRACE, "FDDI - EtherType: %x", 
            hdr->llc.EtherType);
    xTrace2(fddip, TR_FUNCTIONAL_TRACE, "src: %s  dst: %s",
	    fddiHostStr(&hdr->src), fddiHostStr(&hdr->dst));
    xTrace4(fddip, TR_FUNCTIONAL_TRACE, 
            "dsap: %d  ssap: %d  control: %d  org-code: %d",
            hdr->llc.dsap, hdr->llc.ssap, hdr->llc.control,
            (hdr->llc.OrgCodePart1 << (8*sizeof(hdr->llc.OrgCodePart2))) + 
             hdr->llc.OrgCodePart2);
    xIfTrace(fddip, TR_DETAILED) msgShow(msg);
    xAssert(hdr);
    if ( ps->prmSessn ) {
	Msg	pMsg;
	
	xTrace0(fddip, TR_EVENTS,
		"fddiDemux: passing msg to promiscuous session");
	msgConstructCopy(&pMsg, msg);
	xDemux(ps->prmSessn, &pMsg);
	msgDestroy(&pMsg);
    }
#ifdef XK_DEBUG
    /*
     * verify that msg is for this host
     */
    if ( ! (FDDI_ADDRS_EQUAL(hdr->dst, ps->myHost) ||
           FDDI_ADDRESS_ISBROAD(hdr->dst))) {
	xError("fddiDemux: msg is not for this host");
	return XK_FAILURE;
    }

#  if 0
    /* 
     * Temporary for testing
     */
    {
	static int	count = 0;

	/* 
	 * Every 30 packets there is a burst for 10 packets during
	 * which every other packet is delayed.
	 */
	count++;
	if ( ((count / 10) % 3) &&  ! (count % 2) ) {
	    xError("fddiDemux delays packet");
	    Delay(4 * 1000);
	    xError("fddiDemux delay returns");
	} else {
	    xTrace1(fddip, TR_EVENTS, "fddiDemux does not delay (%d)", count);
	}
    }
#  endif
#endif
    actKey.host = hdr->src;
    actKey.type = hdr->llc.EtherType;
    if ( mapResolve(ps->activeMap, &actKey, &s) == XK_FAILURE ) {
	pasKey = actKey.type;
	if ( mapResolve(ps->passiveMap, &pasKey, &e) == XK_SUCCESS ) {
	    xTrace1(fddip, TR_EVENTS,
		    "fddiDemux: openenable exists for msg type %x",
		    ntohs(pasKey));
	    xAssert( ntohs(hdr->llc.EtherType) == relProtNum(e->hlpType,self) );
	    s = fddiCreateSessn(self, e->hlpRcv, e->hlpType, &actKey);
	    if ( s != ERR_XOBJ ) {
		xOpenDone(e->hlpRcv, s, self);
		xTrace0(fddip, TR_EVENTS,
			"fddiDemux: sending message to new session");
	    }
	} else {
	    xTrace1(fddip, TR_EVENTS,
		    "fddiDemux: openenable does not exist for msg type %x",
		    ntohs(pasKey));
	}
    }
    if ( xIsSession(s) ) {
	xPop(s, llp, msg, 0);
    }
    msgDestroy(msg);
    return XK_SUCCESS;
}



static xkern_return_t
fddiClose( s )
    XObj	s;
{
    PState	*ps = (PState *)xMyProtl(s)->state;

    xTrace1(fddip, TR_MAJOR_EVENTS, "FDDI closing session %x", s);
    xAssert( xIsSession( s ) );
    xAssert( s->rcnt <= 0 );
    mapRemoveBinding( ps->activeMap, s->binding );
    xDestroy( s );
    return XK_SUCCESS;
}


static void
demuxStub(ev, arg)
    Event	ev;
    VOID *         arg;
{
    RetBlock	*b = (RetBlock *)arg;

    fddiDemux(b->self, b->llp, &b->msg);
    xFree((char *)arg);
}


static xmsg_handle_t
fddiLoopPush( s, m )
    XObj	s;
    Msg		*m;
{
    RetBlock	*b;

    b = X_NEW(RetBlock);
    msgConstructCopy(&b->msg, m);
    b->hdr = ((SState *)s->state)->hdr;
    msgSetAttr(&b->msg, 0, (VOID *)&b->hdr, sizeof(b->hdr));
    b->self = s->myprotl;
    b->llp = xGetDown(s->myprotl, 0);
    evDetach( evSchedule(demuxStub, b, 0) );
    return XMSG_NULL_HANDLE;
}


static xmsg_handle_t
fddiPush( s, msg )
    XObj	s;
    Msg 	*msg;
{
    xTrace0(fddip, TR_EVENTS, "entered fddiPush");
    msgSetAttr(msg, 0, &((SState *)s->state)->hdr, sizeof(FDDIhdr));
    xPush(xGetDown(s, 0), msg);
    xTrace0(fddip, TR_EVENTS, "leaving fddiPush");
    return XMSG_NULL_HANDLE;
}


static xkern_return_t
fddiPop( s, llp, m, h)
    XObj	s, llp;
    Msg 	*m;
    VOID *      h;
{
    return xDemux(s, m);
} 


static int
fddiControlSessn(s, op, buf, len)
    XObj	s;
    int 	op, len;
    char 	*buf;
{
    SState	*ss = (SState *)s->state;
    
    xAssert(xIsSession(s));
    switch (op) {

      case GETMYHOST:
      case GETMAXPACKET:
      case GETOPTPACKET:
	return fddiControlProtl(xMyProtl(s), op, buf, len);
	
      case GETPEERHOST:
	checkLen(len, sizeof(FDDIhost));
	bcopy((char *) &ss->hdr.dst, buf, sizeof(FDDIhost));
	return (sizeof(FDDIhost));
	
      case GETMYHOSTCOUNT:
      case GETPEERHOSTCOUNT:
	checkLen(len, sizeof(int));
	*(int *)buf = 1;
	return sizeof(int);
	
      case GETMYPROTO:
      case GETPEERPROTO:
	checkLen(len, sizeof(long));
	*(long *) buf = ss->hdr.llc.EtherType;
	return sizeof(long);
	
      case GETPARTICIPANTS:
	{
	    Part	p;

	    partInit(&p, 1);
	    /* 
	     * Remote host
	     */
	    partPush(p, &ss->hdr.dst, sizeof(FDDIhost));
	    return (partExternalize(&p, buf, &len) == XK_FAILURE) ? -1 : len;
	}

      case MAC_SETPROMISCUOUS:
	{
            /*
             * This doesn't seem right (what happens if simfddi is
             * our down protocol?) - fix it LATER!!!
             */
	    PState	*ps = (PState *)xMyProtl(s)->state;
	    checkLen(len, sizeof(int));
	    ps->prmSessn = s;
	    /* 
	     * Tell the device driver to go into promiscuous mode
	     */
	    return xControl(xGetDown(s, 0), op, buf, len);
	}

      case MAC_CLEARPROMISCUOUS:
        {
            xError("MAC_CLEARPROMISCUOUS not yet supported in FDDI");
            return -1;
        }

      default:
	return -1;
    }
}


static int
fddiControlProtl( self, op, buf, len )
    XObj	self;
    int 	op, len;
    char 	*buf;
{
    PState	*ps = (PState *)self->state;

    xAssert(xIsProtocol(self));
  
    switch (op) {

      case GETMAXPACKET:
      case GETOPTPACKET:
	checkLen(len, sizeof(int));
	*(int *) buf = ps->mtu;
	return (sizeof(int));
	
      case GETMYHOST:
	checkLen(len, sizeof(FDDIhost));
	bcopy((char *) &ps->myHost, buf, sizeof(FDDIhost));
	return (sizeof(FDDIhost));

      default:
	return xControl(xGetDown(self, 0), op, buf, len);
  }
  
}


static void
fddiSessnInit(s)
    XObj s;
{
  s->push = fddiPush;
  s->pop = fddiPop;
  s->close = fddiClose;
  s->control = fddiControlSessn;
}


