/*     
 * $RCSfile: psync_support.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: psync_support.c,v $
 * Revision 1.1  1994/04/18 22:02:05  slm
 * Initial revision
 *
 */

/*
 * Abreviations:
 *    Ih = Image of context graph on host h
 *    Vp = participant p's View of context graph
 */

#include "xkernel.h"
#include "psync.h"
#include "psync_internal.h"

void init_nodes( void );
void insert_inviteQ(CID *, PSYNCport, Msg *, XObj);
int  init_graph(PSYNCstate *, Sstate *, PSYNCheader1 *, Msg *, int);
int  update_graph(PSYNCstate *, Sstate *, MID *);
NODE *newnode(Sstate *, MID *, int, Msg *);
NODE *get_node(PSYNCstate *, Sstate *, MID *);
MID  *get_mid(PSYNCstate *, NODE *);
int  get_sender(PSYNCstate *, Sstate *, MID *);
void get_leaves(PSYNCstate *, Sstate *, unsigned char *, MID *);
Msg  *node_to_msg(PSYNCstate *, NODE *);
int  check_path(NODE *, NODE *);
void start_stability_check(PSYNCstate *, Sstate *, MID *);
void compute_indegrees(NODE **, NODE *, Map *, INDEGREE *);
void dumpHdr1(PSYNCheader1 *);
void dumpHdr2(PSYNCheader2 *);
void dumpHdr3(PSYNCheader3 *);
void dumpNodes(NODE *);

static int  redundant_parent(NODE *, NODE *);
static void set_stable_ancestor(NODE *);
static void check_stability(Sstate *, NODE *, unsigned);

/*
 * static void send_restart_msg(XObj);
 * static Msg  get_message(XObj, XObj, MID);
 */

typedef struct node_node {
    struct node_node *next;
    NODE node;
} OUR_NODE;

static OUR_NODE our_node[NUM_NODES];
static OUR_NODE *ofree = &our_node[0];

extern INVITEQ *inviteQ;

extern int tracepsyncp;

void init_nodes()
{
    int i;

    for (i = 0; i < NUM_NODES-1; i++)
        our_node[i].next = &our_node[i+1];
    our_node[NUM_NODES-1].next = NULL;
}

/*
 * Keep in inviteQ the conversation id that has a participant in the local host
 * but hasn't done an openenable.  The first message of the conversation is
 * also stored.
 */
void insert_inviteQ(cid, port, msg, lls)
CID       *cid;
PSYNCport port;
Msg       *msg;
XObj      lls;
{
    INVITEQ *new, *p;

    if ((new = (INVITEQ *)xMalloc(sizeof(INVITEQ))) == NULL)
        Kabort("PSYNC: malloc failed in insert_invitee");
    new->cid.host    = cid->host;
    new->cid.conv_id = cid->conv_id;
    new->port        = port;
    new->msg         = msg;
    new->lls         = lls;
    new->next        = NULL;
    if (inviteQ == NULL) {
        inviteQ = new;
        return;
    }
    for (p = inviteQ; p->next; p = p->next)
	;
    p->next = new;
}

/*
 * Initialize the Ih graph.  Called by psync_push and psync_pop when the first
 * message has been sent/received.
 */
int init_graph(pstate, sstate, hdr1, msg, push)
PSYNCstate   *pstate;
Sstate       *sstate;
PSYNCheader1 *hdr1;
Msg          *msg;
int          push;
{
    NODE *node;

    semWait(&pstate->Ihgraph);
    /* `map_tree' is <iphost,msgid> --> <node> */
    if (mapResolve(sstate->map_tree, (VOID *)&hdr1->mid, (VOID **)&node) !=
	    XK_FAILURE)		/* drop extra copy of the message */
        return (int)node;
    /* Create new message node. */
    node = newnode(sstate, &hdr1->mid, hdr1->sender, msg);
    if (push)			/* called by psync_push */
        sstate->lastmsg[hdr1->sender] = node;
    sstate->first_msg  = FALSE;
    sstate->root       = node;
    sstate->num_leaves = 0;
    /* Bind message to map. */
    mapBind(sstate->map_tree, (VOID *)&hdr1->mid, (VOID *)node);
    semSignal(&pstate->Ihgraph);
    return (int)node;
}

/*
 * Change the paticipant's Vp after some message is received (actually only the
 * leaves are being maintained).  Return 0 if the message is not known to Ih.
 */
int update_graph(pstate, sstate, mid)
PSYNCstate *pstate;
Sstate     *sstate;
MID        *mid;
{
    NODE *node, *parent;
    int  i, j, redundant;

    semWait(&pstate->Ihgraph);
    /* `map_tree' is <iphost,msgid> --> <node> */
    if (mapResolve(sstate->map_tree, (VOID *)mid, (VOID **)&node) ==
	    XK_FAILURE) {
        /* message is not known to Ih */
        semSignal(&pstate->Ihgraph);
	xTrace0(psyncp, TR_FUNCTIONAL_TRACE,
		"PSYNC: exiting update_graph, message not known in Ih");
        return 0;
    }
    xTrace1(psyncp, TR_DETAILED, "PSYNC: update_graph called for node %d",
	    (int)node);

    /*
     * Set links from this node's parents to this node, and update the Vp's
     * leaf pointers by removing the message's parents that WERE leaves.
     */
    for (i = 0; i < node->num_prev; i++) {
	/* for every one of the message's parents */
	parent = node->prev[i];
        parent->next[parent->num_next++] = node;
	for (j = 0; j < sstate->num_leaves; j++) {
	    /* for every leaf of Vp */
	    if (parent == sstate->leaves[j]) {
		/* parent was a leaf, but not any more */
		sstate->leaves[j] = sstate->leaves[--sstate->num_leaves];
		break;
	    }
	}
    }
    sstate->leaves[sstate->num_leaves++] = node;

    /*
     * Eliminate redundant edges.
     * (use do-while loop because test value may change)
     */
    i = 0;
    do {        /* for each parent */
	parent = node->prev[i];
	redundant = FALSE;
	for (j = 0; j < node->num_prev; j++) {  /* for every other parent */
	    if (i != j && redundant_parent(parent, node->prev[j])) {
		redundant = TRUE;
		break;
	    }
	}
	if (redundant) {
	    /*
	     * Remove prev pointer to redundant parent, and remove redundant
	     * parent's next pointer to this node.
	     */
	    node->prev[i] = node->prev[--node->num_prev];
	    for (j = 0; j < parent->num_next; j++) {
		if (parent->next[j] == node) {
		    parent->next[j] = parent->next[--parent->num_next];
		    break;
		}
	    }
	}
	else
	    i++;
    } while (i < node->num_prev);

    semSignal(&pstate->Ihgraph);
    return (int)node;
}

/*
 * Allocate and initialize a new message node.
 */
NODE *newnode(sstate, mid, sender, msg)
Sstate *sstate;
MID    *mid;
int    sender;
Msg    *msg;
{
    NODE *node;

    if (ofree == NULL) {
        if ((node = (NODE *)xMalloc(sizeof(NODE))) == NULL)
            Kabort("PSYNC: malloc failed in newnode");
    }
    else {
        node = &ofree->node;
        ofree = ofree->next;
    }

    node->mid              = *mid;
    node->sender           = sender;
    node->msg              = msg;
    node->recvd            = 0;
    node->num_prev         = 0;
    node->num_next         = 0;
    node->stable_part_mask = ~(~0 << sstate->num_part) &
			     ~(1 << sstate->mypartnum);
    return node;
}

/*
 * Return a message's node; return NULL if message is unknown in Ih.
 */
NODE *get_node(pstate, sstate, mid)
PSYNCstate *pstate;
Sstate     *sstate;
MID        *mid;
{
    NODE *node;

    semWait(&pstate->Ihgraph);
    /* `map_tree' is <iphost,msgid> --> <node> */
    if (mapResolve(sstate->map_tree, (VOID* )mid, (VOID **)&node) ==
	    XK_FAILURE) {
        semSignal(&pstate->Ihgraph);
        return NULL;
    }
    semSignal(&pstate->Ihgraph);
    return node;
}

/*
 * Return a message node's message id.
 */
MID *get_mid(pstate, node)
PSYNCstate *pstate;
NODE       *node;
{
    MID *mid;

    if ((mid = (MID *)xMalloc(sizeof(MID))) == NULL)
	Kabort("PSYNC: malloc failed in get_mid");
    mid = &node->mid;
    return mid;
}

/*
 * Return a message id's message's sender's PSYNC address, or NULL if message
 * is not known in Ih.
 */
int get_sender(pstate, sstate, mid)
PSYNCstate *pstate;
Sstate     *sstate;
MID        *mid;
{
    NODE *node;

    semWait(&pstate->Ihgraph);
    /* `map_tree' is <iphost,msgid> --> <node> */
    if (mapResolve(sstate->map_tree, (VOID *)mid, (VOID **)&node) ==
	    XK_FAILURE) {
        semSignal(&pstate->Ihgraph);
        return -1;
    }
    semSignal(&pstate->Ihgraph);
    return node->sender;
}

/*
 * Build an array of message ids of all the leaves in the participants local
 * view Vp.  Return the number of leaves in num_leaves.
 */
void get_leaves(pstate, sstate, num_leaves, leaves)
PSYNCstate    *pstate;
Sstate        *sstate;
unsigned char *num_leaves;
MID           *leaves;
{
    int i;

    semWait(&pstate->Ihgraph);
    for (i = 0; i < sstate->num_leaves; i++)
        leaves[i] = sstate->leaves[i]->mid;
    *num_leaves = sstate->num_leaves;
    semSignal(&pstate->Ihgraph);
}

/*
 * Return a handle to a message node's message (and increment the reference
 * count).
 */
Msg *node_to_msg(pstate, nd)
PSYNCstate *pstate;
NODE       *nd;
{
    Msg msg;

    msgConstructCopy(&msg, nd->msg);
    return nd->msg;
}

/*
 * Return TRUE if nd1 precedes nd2.
 */
int check_path(nd1, nd2)
NODE *nd1, *nd2;
{
    int  qbegin, qend, i, j;
    NODE *node, *nodes[500];

    qbegin = 0;
    qend   = 1;
    nodes[0] = nd1;
    while (qbegin != qend) {
        node = nodes[qbegin];
        if (node == nd2)
            return TRUE;
        for (i = 0; i < node->num_next; i++) {
            for (j = 0; j < qend; j++)
                if (nodes[j] == node->next[i])
		    break;
	    if (j == qend) {
                nodes[qend] = node->next[i];
                qend++;
                if (qend > 500)
                    Kabort("PSYNC: buffer exceeded in check_path()");
            }
	}
        qbegin++;
    }
    return FALSE;
}

/*
 * Check the stability from node on up.  Ultimately free messages for stable
 * nodes.
 */
void start_stability_check(pstate, sstate, mid)
PSYNCstate *pstate;
Sstate     *sstate;
MID        *mid;
{
    NODE      *node;
    PART_MASK mask;
    int       i;

    node = get_node(pstate, sstate, mid);
    semWait(&pstate->Ihgraph);
    if (node) {
        /*
         * Get the mask for the sender, where the mask is 2**i if sender is
         * part[i] in the converstion's participant list.
         */
        mask = 1 << node->sender;
        /*
         * We don't need to start our check with the current node; its mask
         * should have been initialized to indicate who sent it.
         */
        if (!((node->stable_part_mask &= ~mask) & ~sstate->down_part_mask)) {
            node->stable_part_mask = 0;
#ifdef FREE_STABLE_MESSAGE
            msgDestroy(nd->msg);
#endif
            for (i = 0; i < node->num_prev; i++)
	        set_stable_ancestor(node->prev[i]);
        }
        else
            for (i = 0; i < node->num_prev; i++)
	        check_stability(sstate, node->prev[i], mask);
    }
    else {
        xError("PYSNC: get_node returned NULL in start_stability_check");
	xTrace2(psyncp, TR_DETAILED, "==> mid->msg_id = %d, mid->host = %s",
		mid->msg_id, ipHostStr(&mid->host));
	Kabort("");
    }
    semSignal(&pstate->Ihgraph);
}

void compute_indegrees(depend, nd, indegree_map, que)
NODE     *depend[], *nd;
Map      *indegree_map;
INDEGREE que[];
{
    int  qbegin, qend, i, j, index;
    NODE *node;

    /* `indegree_map' is <node> --> <que_index> */
    mapBind(*indegree_map, (VOID *)&nd, (VOID *)0);
    qbegin = 0;
    qend   = 1;
    que[0].node     = nd;
    que[0].indegree = 0;
    while (qbegin != qend) {
        node = que[qbegin].node;
        for (i = 0; i < node->num_prev; i++) {
	    for (j = 0; depend[j]; j++)
		if (check_path(node->prev[i], depend[j]))
		    break;
	    if (!depend[j]) {  /* no path from node->prev[i] to any depend[j] */
                if (mapResolve(*indegree_map, (VOID *)&node->prev[i],
			       (VOID **)&index) == XK_FAILURE) {
                    mapBind(*indegree_map, (VOID *)&node->prev[i],
			    (VOID *)qend);
		    index = qend;
                    que[index].node     = node->prev[i];
                    que[index].indegree = 1;
                    qend++;
                    if (qend > 350)
                        Kabort("PSYNC: buffer exceeded in compute_indegrees()");
                }
                else
                    que[index].indegree++;
            }
	}
        qbegin++;
    }
}

void dumpHdr1(hdr1)
PSYNCheader1 *hdr1;
{
    int i;

    xIfTrace(psyncp, TR_DETAILED) {
        printf("PSYNC: dumpHdr1: ");
        switch (hdr1->type) {
            case FIRST:
                printf("type=FIRST\n");
	        break;
            case REGULAR:
                printf("type=REGULAR\n");
	        break;
            case REPLY:
                printf("type=REPLY\n");
	        break;
            case RETRANSMIT:
                printf("type=RETRANSMIT\n");
	        break;
            case RESTART:
                printf("type=RESTART\n");
	        break;
            default:
	        printf("unknown type (%d)\n", hdr1->type);
	        break;
        }
        printf("       CID: IPhost = %s, convid = %d\n",
               ipHostStr(&hdr1->cid.host), hdr1->cid.conv_id);
        printf("       MID: host = %s, id = %d, sender = %d\n",
	       ipHostStr(&hdr1->mid.host), hdr1->mid.msg_id, hdr1->sender);
        printf("       part_list: (size = %d)\n", hdr1->num_part);
        for (i = 0; i < hdr1->num_part; i++)
	    printf("         part_list[%d]: port = %d, host = %s\n",
	           i, hdr1->part_list[i].port,
		   ipHostStr(&hdr1->part_list[i].host));
    }
}

void dumpHdr2(hdr2)
PSYNCheader2 *hdr2;
{
    int i;

    xIfTrace(psyncp, TR_DETAILED) {
        printf("PSYNC: dumpHdr2: ");
        switch (hdr2->type) {
            case FIRST:
                printf("type=FIRST\n");
	        break;
            case REGULAR:
                printf("type=REGULAR\n");
	        break;
            case REPLY:
                printf("type=REPLY\n");
	        break;
            case RETRANSMIT:
                printf("type=RETRANSMIT\n");
	        break;
            case RESTART:
                printf("type=RESTART\n");
	        break;
            default:
	        printf("unknown type (%d)\n", hdr2->type);
	        break;
        }
        printf("       CID: IPhost = %s, convid = %d\n",
               ipHostStr(&hdr2->cid.host), hdr2->cid.conv_id);
        printf("       MID: host = %s, id = %d, sender = %d\n",
               ipHostStr(&hdr2->mid.host), hdr2->mid.msg_id, hdr2->sender);
        printf("       depend_list: (size = %d)\n", hdr2->num_depend);
        for (i = 0; i < hdr2->num_depend; i++)
	    printf("         depend_list[%d]: host = %s, id = %d\n",
	           i, ipHostStr(&hdr2->depend_list[i].host),
		   hdr2->depend_list[i].msg_id);
    }
}

void dumpHdr3(hdr3)
PSYNCheader3 *hdr3;
{
    xIfTrace(psyncp, TR_DETAILED) {
        printf("PSYNC: dumpHdr3: ");
        switch (hdr3->type) {
            case FIRST:
                printf("type=FIRST\n");
	        break;
            case REGULAR:
                printf("type=REGULAR\n");
	        break;
            case REPLY:
                printf("type=REPLY\n");
	        break;
            case RETRANSMIT:
                printf("type=RETRANSMIT\n");
	        break;
            case RESTART:
                printf("type=RESTART\n");
	        break;
            default:
	        printf("unknown type (%d)\n", hdr3->type);
	        break;
        }
        printf("       CID: IPhost = %s, convid = %d\n",
	       ipHostStr(&hdr3->cid.host), hdr3->cid.conv_id);
    }
}

void dumpNodes(nd)
NODE *nd;
{
    int i;

    printf("       Node %d, MID host = %s, id = %d, sender = %d\n",
           nd, ipHostStr(&nd->mid.host), nd->mid.msg_id, nd->sender);
    printf(
     "       recvd = %d, num_prev = %d, num_next = %d, stable_part_mask = %d\n",
           (int)nd->recvd, nd->num_prev, nd->num_next, nd->stable_part_mask);

    for (i = 0; i < nd->num_next; i++)
        printf("next[%d] = %d\n", i, nd->next[i]);

    for (i = 0; i < nd->num_next; i++)
        dumpNodes(nd->next[i]);
}

static int redundant_parent(parent1, parent2)
NODE *parent1, *parent2;
{
    int i;

    if (parent1 == parent2)
	return TRUE;
    for (i = 0; i < parent2->num_prev; i++)
	if (redundant_parent(parent1, parent2->prev[i]))
	    return TRUE;
    return FALSE;
}

static void set_stable_ancestor(nd)
NODE *nd;
{
    int i;

    if (nd->stable_part_mask) {
        for (i = 0; i < nd->num_prev; i++)
            set_stable_ancestor(nd->prev[i]);
        nd->stable_part_mask = 0;
#ifdef FREE_STABLE_MESSAGE
        msgDestroy(nd->msg);
#endif
    }
}

/*
 * Check stability from nd on up, masking in mask (to indicate the arrival of a
 * message from a certain participant) into nd's stable_part_mask.
 */
static void check_stability(sstate, nd, mask)
Sstate   *sstate;
NODE     *nd;
unsigned mask;
{
    int i;

    /* Stop looking at ancestors if nd is already stable. */
    if (nd->stable_part_mask) {		/* nd is not yet stable */
        if (~(nd->stable_part_mask) & mask)
            /*
	     * `nd' already has a message from the sender following it so its
	     * arrival doesn't make any difference.
	     */
            return;
	nd->stable_part_mask &= ~mask;
        if (!(nd->stable_part_mask & ~sstate->down_part_mask)) {
            nd->stable_part_mask = 0;
#ifdef FREE_STABLE_MESSAGE
            msgDestroy(nd->msg);
#endif
            for (i = 0; i < nd->num_prev; i++)
	        set_stable_ancestor(nd->prev[i]);
        }
        else {
	    /* Even after masking, nd is not stable.  Need to go to nd's
	     * ancestors and mask in `mask' to indicate the arrival of a
	     * message from another participant.
             */
            for (i = 0; i < nd->num_prev; i++)
	        check_stability(sstate, nd->prev[i], mask);
        }
    }
}

/*
 * static void send_restart_msg(mysessn)
 * XObj mysessn;
 * {
 *     Sstate       *sstate = (Sstate *)mysessn->state;
 *     PSYNCheader3 hdr3;
 *     Msg          msg, tmpmsg;
 *     int          i;
 *
 *     Kabort("PSYNC: send_restart_msg: not supported");
 *     hdr3.type = RESTART;
 *     hdr3.cid  = sstate->cid;
 *
 *     msgConstructBuffer(&msg, (char *)&hdr3, sizeof(PSYNCheader3));
 *     for (i = 0; i < sstate->num_part; i++) {
 *         if (i != sstate->mypartnum) {
 *             if (!xIsObj(xGetDown(mysessn,i))) {
 *                 msgConstructCopy(&tmpmsg, &msg);
 *                 xPush(xGetDown(mysessn,i), &tmpmsg);
 *                 msgDestroy(&tmpmsg);
 *             }
 *         }
 *     }
 *     msgDestroy(&msg);
 * }
 *
 * / *
 *   * Return a handle to a message (and increment the reference count);
 *   * return NULL if message is not known to Ih.
 *   * /
 * static Msg get_message(pstate, sstate, mid)
 * PSYNCstate    *pstate;
 * Sstate        *sstate;
 * MID  mid;
 * {
 *     NODE *node;
 *     Msg  msg;
 *
 *     semWait(&pstate->Ihgraph);
 *     if (mapResolve(sstate->map_tree, (VOID *)&mid, (VOID **)&node) ==
 *             XK_FAILURE) {
 *         semSignal(&pstate->Ihgraph);
 *         msgConstructEmpty(&msg);
 *         return msg;
 *     }
 *     semSignal(&pstate->Ihgraph);
 *     msgConstructCopy(&msg, node->msg);
 *     return *(node->msg);
 * }
 */
