/*
 * $RCSfile: two4one.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: two4one.c,v $
 * Revision 1.4  1996/02/15 21:58:57  mjk
 * added conditional #undef min before #define
 *
 * Revision 1.3  1994/11/10  22:33:49  ho
 * Fixed array size.
 *
 * Revision 1.2  1994/11/08  02:22:23  ho
 * Added xCall interface.
 *
 * Revision 1.1  1994/11/05  03:47:06  ho
 * Initial revision
 *
 * Revision 1.3  1993/12/17  17:46:28  menze
 * Fixed some int-char comparison problems
 *
 */

/*
 *   This is a protocol that simulates asymmetric sessions
 *
 *   It can be used on two modes: to use separate sessions
 *     for sending and receiving ("switch")
 *         or
 *    to use one session for receiving, several for sending ("multiply")
 *      (simulated multicast)
 *
 *   The protocol has a management function in which it is
 *   assisted by its key manager.  In one mode, the mapping from the
 *   "open" and "opendone" addresses to the several outgoing sessions is
 *   done through the key manager.  In the other mode, the mapping from the
 *   "opendone" address to the second (or multiple) addresses to be
 *   used for "xPush" is controlled by the key manager.
 *
 *   The key manager supplies a list of participants to be used
 *   in opening lower protocol 0.
 *
 *   For multicasting, how do we coordinate the several possible openDone's???
 *   Presumably we only want to reply to those participants that we expect
 *   (i.e., the key manager knows about them), and which have either
 *   already contacted us or which we have been able to contact.
 *   Also, must be careful about the problem of trying to send reply
 *   packets while the open's (from opendone) are still pending.
 *
 *   Note, this has some security constraints if it is used to send
 *   information out in both clear and signed forms.  The system
 *   security policy can require that the incoming/outgoing data be
 *   authenticated but not encrypted.  In this case it does not make
 *   sense to accept incoming unauthenticated data.
 *
 */

#include "xkernel.h"
#include "km.h"
#include "two4one.h"
#include "two4one_i.h"

#define savePart(pxx, spxx)	\
    bcopy((char *)(pxx), (char *)(spxx), sizeof(Part) * partLen(p))

/* 
 * Check for a valid participant list
 */
#define partCheck(pxx, name, vsmax, retval)				\
	{								\
	  if ( ! (pxx) || partLen(pxx) < 1 ||				\
	      (partLen(pxx) > vsmax)) { 				\
		xTrace1(two4onep, TR_ERRORS,				\
			"TWO4ONE %s -- bad participants",		\
			(name));  					\
		return (retval);					\
	  }								\
	}

int tracetwo4onep;

#ifdef __STDC__

static XObj		two4oneOpen( XObj, XObj, XObj, Part * );
static xmsg_handle_t	two4onePush(XObj,  Msg *);
static xkern_return_t	two4oneCall(XObj,  Msg *, Msg *);
static xkern_return_t	two4oneOpenDone( XObj, XObj, XObj, XObj );
static xkern_return_t	two4oneOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t	two4oneOpenEnable( XObj, XObj, XObj, Part * );
static int		two4oneControlProtl( XObj, int, char *, int );
static void		two4oneCreateSessn( XObj );
static void		xCopyPart( Part *, Part *, int );
static void		xObjPrint( XObj );
#endif

#define get_partlist(self, lkp, partlist, partbuf, pbsize, xpart) \
    (((lkp=xOpen(self, self, xGetDown(self,KM_PROTL), xpart)) != ERR_XOBJ)\
	&& \
      (xControl(lkp, KM_KEYSIZE, (char *)&pbsize, sizeof(int)), \
       partbuf = xMalloc(pbsize), 1) \
	&& \
      ((pbsize = xControl(lkp, KM_RESOLVE, (char *)partbuf, pbsize)) != -1) \
        && \
      (partlist = (Part *)xMalloc((partExtLen(partbuf)) * sizeof(Part))) \
        && \
      (partInternalize(partlist, partbuf), 1))


#define xGetParticipants(lls, p) \
  xControl(lls, GETPARTICIPANTS, (char *)p, sizeof(p))

#ifdef min
#undef min
#endif
#define min(a, b) ((a)<=(b)?(a):(b))

static void xCopyPart(dst, src, len)
  Part *dst, *src;
{
  int i;

  partInit(dst, len);
  for (i=0; i<min(partLen(src), len); i++)
    {
      *dst = *src;
      dst->len = len;
      dst++; src++;
    }
}   

static void xObjPrint(obj)
     XObj obj;
{
  printf("%s up %s hlptype %s\n",
	 obj->fullName,
	 obj->up->name, obj->hlpType->name);
}

/*
 *  two4oneOpen
 *
 *    we base the service selection entirely on the remote port;
 *    we could use remote<port, ipaddr> instead (e.g., romopt parts 2 )
 *
 */
static XObj
two4oneOpen( self, hlpRcv, hlpType, p )
    XObj	self, hlpRcv, hlpType;
    Part 	*p;
{
    Part	port[2], sp[2], *partlist;
    XObj	lkp = ERR_XOBJ, twosession = ERR_XOBJ;
    signed char	*hostport;
    PState	*ps = (PState *)(self->state);
    char	*partbuf;
    int		pbsize = 500;
    int		port_total;
    XObj	*portlist;
    Part	lpt;
    
    xTrace0(two4onep, TR_MAJOR_EVENTS, "TWO4ONE open");
    partCheck(p, "two4oneOpen", ps->two4one_portsize, ERR_XOBJ);

    xTrace0(two4onep, TR_MAJOR_EVENTS, "TWO4ONE open trying llp(s)");

    /* 
      The service type will be based on the remote port.
      We expect the key manager to map one participant to a list
      of participants.  The remote participant is probably a class D
      IP address.
     */
    savePart(p, sp);
    partInit(port, 1);
    hostport = partPop(sp[REMOTE_PART]);
    partPush(port[0], hostport, ps->two4one_portsize);

    if (hostport && (int)hostport != ANY_PORT && *hostport != (signed char)ANY_PORT
	&&
	get_partlist(self, lkp, partlist, partbuf, pbsize, &port[0]))
      {
	Part *prt;
	int  i;

	port_total = partLen(partlist);
	portlist = (XObj *) xMalloc(port_total * sizeof(XObj));
	for (prt = &partlist[0], i=0; i<port_total; prt++, i++) {
	  xCopyPart(&lpt, prt, 1);
	  if ((portlist[i] = xOpen(self, hlpType, xGetDown(self, TRANSPORT_PROTL), &lpt))
	    != ERR_XOBJ )
	  {
	    xTrace1(two4onep, TR_MAJOR_EVENTS,
		    "TWO4ONE open of llp %d successful", i);
	    xIfTrace(two4onep, TR_MAJOR_EVENTS) { xObjPrint(portlist[i]); }
	  }
	  else {
	    xTrace1(two4onep, TR_ERRORS,
		    "TWO4ONE open of session %d failed", i);
	    xIfTrace(two4onep, TR_MAJOR_EVENTS)
	      { xObjPrint(xGetDown(self, TRANSPORT_PROTL)); }
	    break;
	  }
	}
	xFree((char *)partlist);
	xFree((char *)partbuf);
	if (i>=port_total) {
	  twosession = xCreateSessn(two4oneCreateSessn, hlpRcv, hlpType, self,
				      port_total, portlist);
	  twosession->state =  X_NEW(SState);
	  xFree((char *)portlist);
	}
	else  /* what is correct error termination ? */
	  { xFree((char *)portlist);  return ERR_XOBJ; }
      }
    if (lkp != ERR_XOBJ) xClose(lkp);
    return twosession;
  }

static void
two4oneCreateSessn(self)
    XObj self;
{
    SState	*sstate;
    PState	*pstate;

    pstate = (PState *)self->state;
    sstate = X_NEW(SState);
    self->state = (char *)sstate;
    self->push = two4onePush;
    self->call = two4oneCall;
    self->control = two4oneControlProtl;
  }


/*
 *  two4oneOpenEnable
 *
 *    assume there is only one participant, the convention is that it specifies
 *    the local part ... e.g. the local server port for this machine.
 *
 */

static xkern_return_t
two4oneOpenEnable( self, hlpRcv, hlpType, p )
    XObj	self, hlpRcv, hlpType;
    Part	*p;
{
    xTrace0(two4onep, TR_MAJOR_EVENTS, "TWO4ONE openenable");
    return xOpenEnable( hlpRcv, hlpType, xGetDown(self, TRANSPORT_PROTL), p);
  }

static xkern_return_t
two4oneOpenDisable( self, hlpRcv, hlpType, p )
    XObj self, hlpRcv, hlpType;
    Part *p;
{
    PState	*ps = (PState *)self->state;

    xTrace0(two4onep, TR_MAJOR_EVENTS, "TWO4ONE open disable");
    defaultOpenDisable(ps->passiveMap, hlpRcv, hlpType, p);
    /* need to have record of which llp was openenabled */
    return XK_SUCCESS;
}

/*
 *  two4oneOpenDone
 *
 *
 *
 *
 */
static xkern_return_t
two4oneOpenDone( self, lls, llp, hlpType )
    XObj self, lls, llp, hlpType;
{
    PState	*ps = (PState *)self->state;
    Enable	*e;
    XObj	lkp = ERR_XOBJ;
    Part	p[2], sp[2];
    char	partbuf[sizeof(p)], *partbuf2;
    xkern_return_t	val;
    Part	*partlist;
    int		bufsz;
    
    xTrace0(two4onep, TR_MAJOR_EVENTS, "In TWO4ONE OpenDone");
    if ( mapResolve(ps->passiveMap, &hlpType, &e) == XK_FAILURE ) {
	/* 
	 * This shouldn't happen
	 */
	xTrace0(two4onep, TR_ERRORS,
		"two4oneOpenDone: Couldn't find hlp for new session");
	return XK_FAILURE;
    }

    /* the following section makes decisions based on the first
       component of the local participant; there is as yet no
       support for multiple parts.
    */

    if (xGetParticipants(lls, partbuf)) {
      IPhost	hostport_s; 

          partInternalize(p, partbuf);
	  if (ps->over_ip && ps->type == TWO4ONE_SWITCH) {
	    IPhost *hostport = &hostport_s;
	    /* this is explicitly for IP broadcast (sigh) */
	    xControl(lls, GETMYHOST, (char *)hostport, sizeof(hostport_s));
	    partInit(p, 1);
	    /* change to use the broadcast addr for the lookup */
	    partPush(p[0], hostport, ps->two4one_portsize);
	  }
	  else {
	    char *hostport;

	    savePart(&p, &sp);
	    hostport = partPop(sp[LOCAL_PART]);
	    partInit(p, 1);
	    partPush(p[0], hostport, ps->two4one_portsize);
	  }
	}
    else return XK_FAILURE;
    
    /* verify that the incoming service request is consistent
       with the key manager database */
    if ((get_partlist(self, lkp, partlist, partbuf2, bufsz, p) &&
	 xClose(lkp) == XK_SUCCESS))
      {
	if (ps->type == TWO4ONE_MULTIPLY) {
	  /* open all the relevant lower sessions */
	  Part pl[1], *pb;
	  XObj s;
	  XObj	*downlist = (XObj *)
	    xMalloc(partLen(partlist) * sizeof(XObj));
	  int i;

	  s = xCreateSessn(two4oneCreateSessn, e->hlpRcv, hlpType, self,
			   partLen(partlist), downlist);
	  /* we want to avoid being confused if packets arrive before
	     we have finished; must have the Push routine honor the
	     semaphore
	   */
	  semInit(&((SState *)(s->state))->open_wait, 1);
	  /* now open all the downward sessions */
	  for (i=0, pb=&partlist[0]; i<partLen(partlist); i++) {
	    partInit(pl, 1);
	    *pl = *pb++;
	    pl->len = 1;
	    downlist[i] = xOpen(xGetDown(self, TRANSPORT_PROTL), e->hlpRcv, hlpType, pl);
	    if (downlist[i] == ERR_XOBJ)
	      xIfTrace(two4onep, TR_SOFT_ERRORS) {
		printf("two4onep: failed to open down session %d in opendone",
		       i);
	      }
	  }
	  semSignal(&((SState *)(s->state))->open_wait);
	  xFree((char *)downlist);
	  xFree((char *)partlist);
	  xFree((char *)partbuf2);
	}
	else val = xOpenDone(e->hlpRcv, lls, self);
	return	val;
      }
    return XK_FAILURE;
}

/*
 *  two4oneControlProtl
 *
 */
static int
two4oneControlProtl( self, opcode, buf, len )
    XObj self;
    int opcode, len;
    char *buf;
{
  return xControl(xGetDown(self, TRANSPORT_PROTL), opcode, buf, len);
}

static xmsg_handle_t
two4onePush(s, msg)
    XObj s;
    Msg *msg;
{
    PState	*pstate = (PState *)(s->myprotl->state);
    SState	*sstate;
    int i;
    xkern_return_t val = XK_FAILURE;
    
    xTrace0(two4onep, TR_EVENTS, "two4one push");
    sstate = (SState *)s->state;

    /* simulated multicast */
    if (pstate->type == TWO4ONE_MULTIPLY) {
      for (i=0; i<s->numdown; i++) {
	    xTrace1(two4onep, TR_MORE_EVENTS,
		"two4one_push: pushing to session[%d]",i);
	    xAssert(xIsSession(xGetDown(s, i)));
	    if ((xGetDown(s, i) != ERR_XOBJ) && 
		((val = xPush(xGetDown(s, i), msg)) != XK_SUCCESS))
	      xTrace1(two4onep, TR_SOFT_ERRORS, 
		"two4one_push: failed to session[%d]",i);
        }
    }

    /* asymmetric sessions */
    if (pstate->type == TWO4ONE_SWITCH) {
      for (i=0; i<s->numdown; i++) {
	if (sstate->pushdown[i]) {
	  xTrace1(two4onep, TR_MORE_EVENTS,
		  "two4one_push: pushing to session[%d]",i);
	  xAssert(xIsSession(xGetDown(s, i)));
	  if ((val = xPush(xGetDown(s, i), msg)) != XK_SUCCESS)
	    xTrace1(two4onep, TR_SOFT_ERRORS, 
		    "two4one_push: failed to session[%d]",i);
	}
      }
    }
    return val;
  }

static xkern_return_t
two4oneCall(s, msg, rmsg)
    XObj s;
    Msg *msg, *rmsg;
{
    PState	*pstate = (PState *)(s->myprotl->state);
    SState	*sstate;
    int i;
    xkern_return_t val = XK_FAILURE;
    
    xTrace0(two4onep, TR_EVENTS, "two4one call");
    sstate = (SState *)s->state;

    /* simulated multicast */
    if (pstate->type == TWO4ONE_MULTIPLY) {
      for (i=0; i<s->numdown; i++) {
	    xTrace1(two4onep, TR_MORE_EVENTS,
		"two4one_call: pushing to session[%d]",i);
	    xAssert(xIsSession(xGetDown(s, i)));
	    if ((xGetDown(s, i) != ERR_XOBJ) && 
		((val = xCall(xGetDown(s, i), msg, rmsg)) != XK_SUCCESS))
	      xTrace1(two4onep, TR_SOFT_ERRORS, 
		"two4one_call: failed to session[%d]",i);
        }
    }

    /* asymmetric sessions */
    if (pstate->type == TWO4ONE_SWITCH) {
      xTrace0(two4onep, TR_SOFT_ERRORS, "two4one: switch option not sensible with xCall interface");
      val = XK_FAILURE;
    }
    return val;
  }

static xkern_return_t romfindparts(XObj, char **, int, int, void *);
static xkern_return_t rommultiply(XObj, char **, int, int, void *);
static xkern_return_t romoverip(XObj, char **, int, int, void *);
static XObjRomOpt rmopts[3] = {
			       { "partcount", 2, romfindparts },
			       { "multiply", 1, rommultiply },
			       { "overip", 1, romoverip }
			     };


static xkern_return_t
romfindparts(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  int i;

  if (fields > 1 && (i= atoi(ptr[2]))>=0)
    ((PState *)(self->state))->popparts = i;
  return XK_SUCCESS;
}

static xkern_return_t
romoverip(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  ((PState *)(self->state))->over_ip = 1;
  return XK_SUCCESS;
}

static xkern_return_t
rommultiply(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  ((PState *)(self->state))->type = TWO4ONE_MULTIPLY;
  xTrace0(two4onep, TR_MAJOR_EVENTS, "two4one type MULTIPLY");
  return XK_SUCCESS;
}


void
two4one_init( self )
    XObj self;
{
    int		i;
    PState	*ps;
    char	buffer[TWO4ONE_MAX_PORTSIZE];

    xTrace0(two4onep, TR_GROSS_EVENTS, "TWO4ONE init");
    xAssert(xIsProtocol(self));
    ps = X_NEW(PState);
    self->state = (VOID *)ps;
    ps->two4one_portsize = xControl(xGetDown(self, KM_PROTL),
				    KM_HOSTSIZE, buffer, sizeof(buffer));
    if (ps->two4one_portsize < 1) {
      xTrace1(two4onep, TR_ERRORS, "two4one init: Invalid key size %d", ps->two4one_portsize);
      return ;
    }
    for ( i=0; i < self->numdown; i++ ) {
	XObj	llp;

	llp = xGetDown(self, i);
	if ( ! xIsProtocol(llp) ) {
	    xTrace1(two4onep, TR_ERRORS, "TWO4ONE can't find protocol %d", i);
	    xError("TWO4ONE init errors -- not initializing");
	    return;
	} else {
	    xTrace2(two4onep, TR_GROSS_EVENTS, "TWO4ONE llp %d is %s", i, llp->name);
	}
    }
    ps->passiveMap = mapCreate(TWO4ONE_MAP_SZ, ps->two4one_portsize); 
    ps->popparts = 1;
    xTrace0(two4onep, TR_MAJOR_EVENTS, "two4one default participants 1");
    ps->type = TWO4ONE_SWITCH;
    xTrace0(two4onep, TR_MAJOR_EVENTS, "two4one default type SWITCH");

    self->control = two4oneControlProtl;
    self->open = two4oneOpen;
    self->openenable = two4oneOpenEnable;
    self->opendisable = two4oneOpenDisable;
    self->opendone = two4oneOpenDone;
    findXObjRomOpts(self, rmopts, 3, 0);
}
